;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 01_stacks_1.lisp - Zásobníkové programování
;;;
;;; Jazyk pro myšlený zásobníkový stroj.
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; fáze 1 - základy 
;;;

;;
;; Hlavní zásobníky
;;
 
;; Datový zásobník (Result Stack) bude v proměnné *rslt*
;; Programový zásobník (Execution Stack) bude v proměnné *exec*
;; Kvůli pohodlnějšímu ladění budou zastíněny v hlavní funkci execute.
;; (Aby byly zásobníky snadno vidět ve stepperu.)
 
(defvar *rslt*)
(defvar *exec*)

;; Ještě zásobník, který budeme potřebovat příště. Definujeme ho tady, 
;; abychom nemuseli příště upravovat funkci execute
(defvar *ret*)


;; Slovník je zásobník párů (klíč . hodnota)
;; práce se slovníky:

(defun dict-find (key dict)
  (let ((pair (find key dict :key #'car)))
    (unless pair (error "Key not found in dictionary: ~s" key))
    (cdr pair)))

(defmacro dict-push (key val dict)
  `(push (cons ,key ,val) ,dict))

;; Slovník definic slov. Každá definice je pár (slovo . kód).
;; Kód je buď lispová funkce (vestavěné slovo), nebo kód našeho jazyka
;; v seznamu (uživatelsky definované slovo).

(defvar *word*)
(setf *word* '())

;; 
;; Slova
;;

;; Slova jsou lispovské klíče
(defun wordp (el)
  (keywordp el))

(defun word-code (w)
  (dict-find w *word*))

;; Vykonání slova (je-li na vrcholu programového zásobníku)
(defun exec-word (word)
  (let ((code (word-code word))) 
    (if (functionp code)
        (built-in-exec code)
      (user-exec code))))

(defun built-in-exec (fun)
  (funcall fun))

(defun user-exec (uc)
  (setf *exec* (append uc *exec*)))

;;
;; Definice slova:
;;

(defun define-word (name code)
  (dict-push name code *word*))

;; Makro defprim používáme na definici vestavěného slova (primitiva)
;; Primitivum vždy vykonává vedlejší efekt na zásobnících *exec* a *rslt*
;; (později i dalších). Jeho návratová hodnota není použita.
;; V makru defprim uvádíme kód (tělo), makro z něj vytvoří funkci bez
;; parametru a tu uloží do slovníku.
(defmacro defprim (name &body body)
  `(define-word ',name (lambda () ,@body)))

;;
;; Základní vestavěná slova:
;;

;; (a b -- b a) 
(defprim :swap
  (let ((b (pop *rslt*))
        (a (pop *rslt*)))
    (setf *rslt* (append `(,a ,b) *rslt*))))

;; (a b c -- b c a) 
(defprim :rot
  (let ((c (pop *rslt*))
        (b (pop *rslt*))
        (a (pop *rslt*)))
    (setf *rslt* (append `(,a ,c ,b) *rslt*))))

;; (a -- )
(defprim :drop 
  (pop *rslt*))

;; (a -- a a)
(defprim :dup
  (push (car *rslt*) *rslt*))

;; (a b -- a b a)
(defprim :over
  (push (cadr *rslt*) *rslt*))

;; Funkce word-bin-op je pomůcka na zjednodušení definic slov, která
;; odebírají dvě hodnoty z datového zásobníku a vracejí na něj výsledek.
(defun word-bin-op (fun)
  (let ((b (pop *rslt*))
        (a (pop *rslt*)))
    (push (funcall fun a b) *rslt*)))

;; (a b -- součet a + b)
(defprim :+ 
  (word-bin-op #'+))

;; (a b -- rozdíl a - b)
(defprim :- 
  (word-bin-op #'-))

;; (a b -- součin a * b)
(defprim :*
  (word-bin-op #'*))

;; (a b -- podíl a / b)
(defprim :/
  (word-bin-op #'/))

;; (a b -- log. hodnota a = b)
(defprim :=
  (word-bin-op #'=))

;; Obecné porovnávání:
(defprim :eql
  (word-bin-op #'eql))

;; Slovo :if používá pomocný zásobník
(defvar *seek*)

(defprim :if
  (when (pop *rslt*)
    (push :seek *seek*)
    (push :seek *exec*)))

(defprim :else
  (push :seek *seek*)
  (push :seek *exec*))

(defprim :then
  nil)

(defprim :seek
  (let ((top (pop *exec*)))
    (cond ((eql top :if) 
           (push :skip *seek*) 
           (push :skip *exec*))
          ((find top '(:else :then))
           (pop *seek*))
          (t (push :seek *exec*)))))

(defprim :skip
  (let ((top (pop *exec*)))
    (cond ((eql top :if) 
           (push :skip *seek*) 
           (push :skip *exec*))
          ((eql top :then)
           (pop *seek*)
           (push (car *seek*) *exec*))
          (t (push :skip *exec*)))))

;; Slovo na definici uživatelského slova
;; (jako jediné funguje prefixově)
(defprim :def 
  (let ((word (pop *exec*)))
    (define-word word *exec*)
    (setf *exec* '())
    (push word *rslt*)))

;;
;; Spuštění programu.
;;

;; Zpracování obecné hodnoty. Přesouvá se na datový zásobník.
;; (Funkce exec-word na zpracování slova je už definována nahoře)
(defun exec-val (elem)
  (push elem *rslt*))

(defun exec-elem (elem)
  (if (wordp elem) 
      (exec-word elem)
    (exec-val elem)))

;; Tady vytváříme vazby proměnných *ret*, *seek*, *rslt*, *exec*.
;; Vazby budou dynamické.
(defun execute (&rest code)
  (let ((*ret* '())
        (*seek* '())
        (*rslt* '())
        (*exec* code))
    (loop (when (null *exec*) (return))
          (exec-elem (pop *exec*)))
    (pop *rslt*)))

#|

;; Výrazy je dobré vyhodnocovat v Listeneru a ne tady přes F8.
;; Když se rekurzivní výraz zacyklí, je pak snadné ho zastavit tlačítkem Break.
;; Vždy je dobré krokovat přes breakpoint, jak jsem ukazoval na přednášce.
;; Tak vidíte, jak se postupně vykonávají jednotlivé kroky výpočtu
;; a můžete sledovat stav zásobníků.

(execute 1)
(execute 3 2 :- 4 :*)

(execute 1 2 :swap :drop)

;; Výpočet 4*(4 + 5):
(execute 4 5 :over :+ :*)

(execute 0 1 := :if 2 5 :* :else 2 10 :* :then)
(execute 1 1 := :if 2 5 :* :else 2 10 :* :then)

(execute 1 1 := :if 1 0 := :if 2 :else 3 :then :else 1 0 := :if 4 :else 5 :then :then)
(execute 1 1 := :if 1 0 := :if 2 :else 3 :then :else 1 1 := :if 4 :else 5 :then :then)

;; Definice nového slova v jazyce:
(execute :def :sqr :dup :*)

(execute 5 :sqr)

|#



#|

(execute :def :fact
              :dup 0 := :if :dup 1 :- :fact :* :else
                            :drop 1 :then)

(execute 5 :fact)

;; Iterativní faktoriál. Porovnejte, co se děje se zásobníkem proti
;; předchozí verzi.

;; pomocné slovo :factiter (vzpomeňte si, že u iterativní rekurze jsme
;; obvykle potřebovali napsat pomocnou funkci)
;; Na vrcholu zásobníku očekává už vypočítaný součin, pod ním číslo
;; udávající, kolikrát je ještě třeba násobit.

(execute :def :factiter 
              :over :dup 1 := :if :* :swap 1 :- :swap :factiter :else
                                  :drop :swap :drop :then)

(execute :def :ifact 1 :factiter)

(execute 5 :ifact)

|#

;; Šikovná pomůcka (a další ukázka uživatelsky definovaného slova): 
;; rotace opačným směrem.

(execute :def :rot- :rot :rot)

#|

;; n-té Fibonacciho číslo počínaje n = 1

(execute :def :fib 0 1 :rot :xfib)

(execute :def :xfib
              :dup 1 := :if :rot- :dup :rot :+ :rot 1 :- :xfib :else
                            :drop :swap :drop :then)

(execute 10 :fib)

|#

