;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 03_atomic.lisp - Složené atomické operace
;;;

;;
;; Optimalizace přepsáním maker na speciální operátory
;;

;; Procedura na porovnávání
(execute '(:eql :ret) 'eql? :bind)

;; Procedura na odstranění prvku seznamu
(evaluate '(define remove (lambda (item lst)
                            (if lst
                                (if (eql? (car lst) item)
                                    (remove item (cdr lst))
                                  (cons (car lst)
                                        (remove item (cdr lst))))))))

;; Odkomentováním následující části dojde k nahrazení makra dotimes
;; za speciální operátor dotimes 
#|

;; dotimes přestane být makro
(evaluate '(set! macros (remove 'dotimes macros)))

;; dotimes začne být speciálním operátorem
(evaluate '(set! specials (cons 'dotimes specials)))

;; Procedura vykonávající dotimes.
;; Vstup: ((i n-expr) b1 ... bn)
(execute '(:split begin :swap :cons        ; (begin b1 ... bn) (i n-expr)
           :swap :split :split :drop       ; n-expr i (begin b1 ... bn)
           eval :val :clsub                ; n i (begin b1 ... bn)
           :swap :dup 0 :swap :bind        ; i n (begin b1 ... bn)
           :lbl
           :dup :val :rot           ; n i-val i (begin b1 ... bn)
           :dup :rot-               ; n i-val n i (begin b1 ... bn) 
           :<                       ; nil/t n i  (begin b1 ... bn)
           :if 
               nil
           :else 
               :rot                        ; (begin b1 ... bn) n i
               :dup eval :val :clsub :drop
               :rot-                       ; n i (begin b1 ... bn)
               :swap                       ; i n (begin b1 ... bn)
               :dup :val 1 :+              ; i-val+1 i n (begin b1 ... bn)
               :over :set! :drop           ; i n (begin b1 ... bn)
               t
           :then
           :cjmp
           :drop :drop :drop nil :ret
           

) 'dotimes :set!)
|#

#|
;; Scheme

(dotimes (i 10)
  (dotimes (j 10)
    (print (list i j))))
|#


;; Přepsání makra list na speciální operátor 
#|
(evaluate '(set! macros (remove 'list macros)))
(evaluate '(set! specials (cons 'list specials)))
(execute '(eval :val :swap mapcar :val :clsub :ret) 'list :set!)
|#

;;
;; Změna definice makra co
;;
;; Procesy vytvořené makrem budou nově čekat, než se vytvoří všechny procesy. 
;;

;; Pomocné makro:
(evaluate '(defmac xco (lambda (args)
                       (if (rest args)
                           (let ((endp (gensym "ENDP"))
                                 (arg (gensym "ARG")))
                             (list 'let (list (list endp nil))
                                   (list 'process-run-procedure 
                                         (list 'lambda (list arg)
                                               (list 'await (first args))
                                               (second args)
                                               (list 'set! endp t))
                                         nil)
                                   (cons 'xco (cons (first args) (rest (rest args))))
                                   (list 'await endp)))
                         (list 'set! (first args) t)))))   

;; Samotné makro:
(evaluate '(set! co (lambda (args)
                      (let ((start (gensym "START")))
                        (list 'let (list (list start nil))
                              (cons 'xco (cons start args)))))))

#|
;; Scheme

(co 
 (print 1)
 (print 2))
|#

;;
;; Logika
;;

;; Funkce not
(evaluate '(define not (lambda (val)
                         (if val nil t))))

;; Makro and
(evaluate '(defmac and (lambda (args)
                        (if args
                            (if (cdr args)
                                (list 'if (first args)
                                      (cons 'and (rest args)))
                              (first args))
                          t))))

#|
;; Scheme

(and)
(and 1)
(and 1 2 3)
|#

;; Makro or
(evaluate '(defmac or (lambda (args)
                        (if args
                            (if (cdr args)
                                (let ((result (gensym "RESULT")))
                                  (list 'let (list (list result (first args)))
                                        (list 'if result result (cons 'or (rest args)))))
                              (first args))
                          nil))))

#|
;; Scheme

(or)
(or 1)
(or nil 1 2)
|#

;; Globální proměnná *atomic* rozhoduje, zda je kód prováděn atomicky.
(defvar *atomic* nil)

;; Upravíme execute tak, aby nemohlo dojít k přerušení vykonávání
;; v případě, že je *atomic* Pravda.
(setf dspec:*redefinition-action* :quiet)
(defun execute (&rest code)
  (let ((*ret* '()) 
        (*seek* '())
        (*rslt* '())
        (*exec* code) 
        (*time* 0)
        (*atomic* nil)
        (*rprs-head* (list (make-idle-process)))
        (*rprs-tail* '()))
    (loop 
       (cond 
        ((null *exec*)
         (return))
        ((= *time* 0)
         (timer-handler))
        (t
         (exec-elem (pop *exec*))
         ;; Při provádění složeného atomického příkazu 
         ;; zastavíme čas.
         (unless *atomic*
           (decf *time*)))))
    (pop *rslt*)))

(setf dspec:*redefinition-action* :warn)

;;; Slova uvozující a ukončující atomický kód.
;; ( -- )
(defprim :aon
  (setf *atomic* t))

;; ( -- )
(defprim :aoff 
  (setf *atomic* nil))

#|
0 n :bind :aon n :val 1 :+ n :set! :aoff n :val :unbind 
|#

;; Speciální operátor atomic
;; Syntax: 
;; (atomic body1 ... bodyn)
;;
;; Tělo se vyhodnotí atomicky.
;; Atomické kódy není možné vnořovat.

(execute 'atomic 'specials :val :cons 'specials :set!)
(execute '(begin :swap :cons eval :val :aon :clsub :aoff :ret) 'atomic :bind)


#|
;; Scheme

;; Atomická inkrementace počítadla:
(let ((n 0))
  (co
   (dotimes (i 100)
     (atomic (set! n (+ n 1))))
   (dotimes (i 100)
     (atomic (set! n (+ n 1)))))
  n)
|#


;; Generování náhodných čísel.
(defprim :rnd ; ( num -- náhdoné číslo do `num`)
  (push (random (pop *rslt*)) *rslt*))

#|
10 :random
|#

;; Pozastavení běhu programu.
(define-word :delay ; (n --)
             '(:lbl 1 :- :dup 0 :> :cjmp :drop))

#|
1000000 :delay
|#


;; Slovo na rozpoznání atomického kódu.
;; ( -- logická hodnota )
(defprim :a? 
  (push *atomic* *rslt*))


;; Změníme chování slova :await tak, aby v případě,
;; že podmínka není splněna a kód je vykonáván atomicky
;; se na nějaký čas z atomického vykonávání ustoupilo.

(execute
 :def :await
 :lbl
 :dup :clsub 
 :dup :if :a? :if :else :aoff 100 :rnd :delay :aon :then 
      :else 
      :then
 :not :cjmp :drop)



;; Test zásobníkového výpočtu používající upravený :await
#|
;; Stacks
nil end? :bind 
nil start :bind
nil (:aon (start :val :ret) :await t end? :set! :aoff :quit) :prsub
100 :delay
t start :set! 
(end? :val :ret) :await
:unbind
:unbind
|#

#|
;; Scheme

;; Může vytisknout 0, 1, -1.
(let ((n 1))
  (co
   (atomic 
    (set! n (- n 1))
    (print n))
   (atomic 
    (set! n (- n 1)))
   (atomic 
    (set! n (+ n 1)))))

;; Nevytiskne záporné číslo:
(let ((n 1))
  (co
   (atomic 
    (await (< 0 n))
    (set! n (- n 1))
    (print n))
   (atomic 
    (set! n (- n 1)))
   (atomic 
    (set! n (+ n 1)))))

;; Makro await může způsobit, že se kód nevykoná atomicky:
(let ((n 0))
  (co
   (atomic
    (set! n 1)
    (await (= n 2))
    (print "Není atomické"))
   (set! n 2)))

;; Speciální operátor await může být pouze první argument speciálního operátoru atomic.
|#

;; Přidáme syntax: [expr1 ... exprn]
;; Výrazy v hranatých závorkách se vyhodnotí atomicky.

(defun left-brack-reader (stream char)
  (declare (ignore char))
  `(atomic ,@(read-delimited-list #\] stream t)))

(defun right-paren-reader (stream char)
  (declare (ignore stream))
  (error "Non-balanced ~s encountered." char))

(set-macro-character #\[ #'left-brack-reader) 
(set-macro-character #\] #'right-paren-reader)

#|
;; Scheme
(let ((n 0))
  (co
   (dotimes (i 100)
     [(set! n (+ n 1))])
   (dotimes (i 100)
     [(set! n (+ n 1))]))
  n)
|#


;;
;; Tvrzení
;;

;; Přidáme slovo :assert-error na vyvolání chyby při nesplnění výjimky.

(defprim :assert-error
  (error "The assertion ~A failed." (pop *rslt*)))


;; Speciální operátor assert
;; Syntax: (assert cond)
;;
;; Pokud není podmínka cond splněna, vyvolá chybu.

(execute 'assert 'specials :val :cons 'specials :set!)
(execute '(:split :drop :dup eval :val :clsub :if :assert-error :else :drop nil :then :ret) 'assert :bind)

#|
Scheme:
(let ((n 0))
  (assert (= n 1)))

(let ((n 0))
  (co 
   (begin
    (set! n 1)
    {(= n 1)})
   (begin
    (set! n 2))))
             
|#




;; Přidáme syntax: {cond}
;; Výraz je ekvivalentní výrazu (assert cond) 

(defun left-brace-reader (stream char)
  (declare (ignore char))
  `(assert ,(car (read-delimited-list #\} stream t))))

(set-macro-character #\{ #'left-brace-reader) 
(set-macro-character #\} #'right-paren-reader)

;; Zařídíme, aby editor rozuměl hranatým a složeným závorkám
(editor::set-vector-value
 (slot-value editor::*default-syntax-table* 'editor::table) '(#\[ #\{) 2)
(editor::set-vector-value
 (slot-value editor::*default-syntax-table* 'editor::table) '(#\] #\}) 3)



#|
;; Scheme:

(let ((n 0))
  (co
   (dotimes (i 100)
     [(await (< 0 n))
      (set! n (- n 1))
      (print (list 0 i n))]
     {(<= 0 n)})
   (dotimes (i 100)
     [(set! n (+ n 1))
      (print (list 1 i n))])))

;; Procesy tisknou: (process i n)
|#