;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 06_vectors_and_lists.lisp - Vektory a seznamy
;;;
;;; Od této verze není interpret Scheamu dále udržovaný. 
;;; Je tedy potřeba používat kompilátor. Například: (compiled-scheme-repl)
;;;

;; Podobně jako v lispu macroexpand-1
;; očekává výraz a pokud se jedná o složený výraz,
;; kde operátor je makro, provede expanzi makra.
;; Jinak vrací původní výraz.
;; Funkce je užitečná na testování expanze maker.
(defun scheme-macroexpand-1 (form)
  (if (consp form)
      (let ((macro (find-macro (car form))))
        (if macro
            (apply macro (cdr form))
          form))
    form))

#|
(scheme-macroexpand-1 '(and a b))
|#
  
;; Nejprve si přidáme pár užitečných operátorů.

;; Modulo
;; (a b -- a modulo b)
(defprim :mod 
  (let ((val (pop *rslt*)))
    (push
     (mod (pop *rslt*) val)
     *rslt*)))

#|
(execute 5 2 :mod)
|#

(execute  '(:mod :ret) 'mod :bind)

#|
(compiled-evaluate '(mod 5 2))
|#


;; Procedura na čekání
(compiled-evaluate '(define sleep (lambda (n)
                                    (dotimes (i n)
                                      nil))))

#|
;; Scheme

(co 
 (begin
  (sleep 10)
  (print 1))
 (print 2))
|#

;; Makra na inkrementaci a dekrementaci proměnné.
(def-scheme-macro inc! (var)
  `(set! ,var (+ ,var 1)))

(def-scheme-macro dec! (var)
  `(set! ,var (- ,var 1)))

#|
;; Scheme

(let ((a 1) (b 1))
  (inc! a)
  (dec! b)
  (print a)
  (print b))
|#


;; Vektor je jednorozměrné pole.
;;
;; Práce s vektory v Lispu:
;; - Literál: #(el1 ... eln)
;; - Funkce na vytvoření: make-array
;; - Místo na prvek vektoru: (aref vector index)
;; - zjištění délky: array-dimension


#|
;; Scheme

#(1 2 3)
|#

;; Vestavěná slova na práci s vektory:
;; ( length -- vector)
(defprim :vect
  (push
   (make-array (list (pop *rslt*)))
   *rslt*))

;; ( vector index -- element)
(defprim :vref 
  (let ((index (pop *rslt*))
        (vector (pop *rslt*)))
    (push (aref vector index) *rslt*)))

;; ( vector value index -- value)
(defprim :vset!
  (let ((index (pop *rslt*))
        (value (pop *rslt*))
        (vector (pop *rslt*)))
  (setf (aref vector index) value)
  (push value *rslt*)))

;; ( vector -- length)
(defprim :vlen
  (push (array-dimension (pop *rslt*) 0) *rslt*))

#|
3 :vect 

#(1 2 3) 1 :vref

#(1 2 3) :dup 4  1 :vset! :drop

#(1 2 3) :vlen
|#

;; Procedury na práci s vektory:
(execute '(:vect :ret) 'make-vector :bind)

(execute '(:vref :ret) 'vref :bind)

(execute '(:swap :vset! :ret) 'vset! :bind)

(execute '(:vlen :ret) 'vlen :bind)

#|
(let ((v1 (make-vector 5)))
  (print v1)
  (dotimes (i 5)
    (vset! v1 i (* i i)))
  (dotimes (i 5)
    (print (vref v1 i)))
  v1)
|#


;; Makro vect slouži k pohodlnému vytváření vektorů.
;; Syntax: (vect length expr)
;; Makro vytvoří vektor délky length, 
;; kde každý prvek vznikne vyhodnocením výrazu expr.
 
(def-scheme-macro vect (length-expr expr)
  (let ((length (gensym "LENGTH"))
        (vector (gensym "VECTOR"))
        (i (gensym "I")))
    `(let ((,length ,length-expr))
       (let ((,vector (make-vector ,length)))
         (dotimes (,i ,length)
           (vset! ,vector ,i ,expr))
         ,vector))))

#|
;; Scheme

;; Vektor semaforů:
(vect 5 (sem 1))
|#



;; Marko na vytváření vektoru ze zadaných prvků.
(defun vector-compile-set (vector args index)
  (when args
    `((vset! ,vector ,index ,(car args))
      ,@(vector-compile-set vector (cdr args) (+ index 1)))))

(def-scheme-macro vector (&rest args)
  (let ((vector (gensym "VECTOR")))
    `(let ((,vector (make-vector ,(length args))))
       ,@(vector-compile-set vector args 0)
       ,vector)))

#|
;; Scheme

(vector (+ 1 2) (* 2 2) 4)
|#

;;
;; Seznamy
;;

;; Přidáme slova na změnu car a cdr párů.

;; ( pair value -- value)
(defprim :set-car!
  (let ((value (pop *rslt*))
        (pair (pop *rslt*)))
    (setf (car pair) value)
    (push value *rslt*)))

;; ( pair value -- value)
(defprim :set-cdr!
  (let ((value (pop *rslt*))
        (pair (pop *rslt*)))
    (setf (cdr pair) value)
    (push value *rslt*)))

#|
;; Stacks

(1 . 2) :dup 3 :set-car! :drop

(1 . 2) :dup 3 :set-cdr! :drop
|#

;; Procedury na změnu car a cdr páru.
(execute '(:set-car! :ret) 'set-car! :bind)
(execute '(:set-cdr! :ret) 'set-cdr! :bind)

#|
;; Scheme
(let ((a (cons 1 2)))
  (set-car! a 3)
  (set-cdr! a 4)
  a)
|#

;; Procedruy lref a lset! zjišťují a mění prvek seznamu na indexu.
(compiled-evaluate '(define lref (lambda (lst i)
                                   (if (= i 0)
                                       (car lst)
                                     (lref (cdr lst) (- i 1))))))


(compiled-evaluate '(define lset! (lambda (lst i val)
                                   (if (= i 0)
                                       (set-car! lst val)
                                     (lset! (cdr lst) (- i 1) val)))))

#|
(let ((lst (list 1 2 3)))
  (lset! lst 1 5)
  (print (lref lst 1))
  l)
|#

;; Procedury na změnu prvních pěti prvků seznamu.
(compiled-evaluate '(define set-first! set-car!))

(compiled-evaluate '(define set-second! (lambda (lst val)
                                          (lset! lst 1 val))))

(compiled-evaluate '(define set-third! (lambda (lst val)
                                          (lset! lst 2 val))))

(compiled-evaluate '(define set-fourth! (lambda (lst val)
                                          (lset! lst 3 val))))

(compiled-evaluate '(define set-fifth! (lambda (lst val)
                                          (lset! lst 4 val))))




;; Makro co-dotimes
;; Syntax:
;;   (co-dotimes (var form) . body)
;;
;; Funguje podobně jako makro dotimes, ale každou iteraci vyhodnocuje paralelně
;; v různých procesech. Čeká, až všechny procesy skončí.

(def-scheme-macro co-dotimes (var-form &rest body)
  (let ((var (car var-form))
        (count-form (cadr var-form))
        (fun (gensym "FUN"))
        (count (gensym "COUNT"))
        ;; Semafory zařídíme start i čekání na skončení procesů:
        (start (gensym "START"))
        (end (gensym "END"))
        (i (gensym "I")))
    `(let ((,start (sem 0))
           (,end (sem 0))
           ;; Funkce, která se bude volat paralelně:
           (,fun (lambda (,var)
                   (p ,start)
                   ,@body 
                   (v ,end)))
           (,count ,count-form))
       ;; Spuštění procesů:
       (dotimes (,i ,count)
         (process-run-procedure ,fun ,i))
       ;; Start vykonávání procesů:
       (dotimes (i ,count)
         (v ,start))
       ;; Čekání, až procesy skončí:
       (dotimes (i ,count)
         (p ,end)))))

;; Nastavení odsazení operátoru co-dotimes
(editor:setup-indent "co-dotimes" 1)

#|
;; Scheme

;; Paralelně desetkrát vytiskne čísla od nuly po deset:
(co-dotimes (i 10)
  (dotimes (j 10)
    (print i)))


;; V deseti procesech stokrát inkrementujeme počítadlo:
(let ((n 0)
      (l (lock)))
  (co-dotimes (i 10)
    (dotimes (j 100)
      (with-lock (l)
        (set! n (+ n 1)))))
  n)
|#
